from pprint import pprint  # pylint: disable=unused-import
from collections import OrderedDict
from typing import (
    Union,
    Optional,
    Tuple,
    Mapping,
    MutableMapping,
    List,
    cast,
    Iterable,
)

__all__ = ["HttpHeaderStorage"]


HttpHeaderStream = Union[
    Mapping[str, str],
    Iterable[Tuple[str, str]],
]


class HttpHeaderStorage(object):
    """Storage for HTTP Headers.

    The storage maps string keys to one or multiple string values.
    Keys are case insensitive though the original case is stored.
    """

    def __init__(
        self, data: Optional[HttpHeaderStream] = None, charset: str = "utf-8"
    ) -> None:
        self._store: MutableMapping[str, List[str]] = OrderedDict()
        self._charset = charset
        if data is not None:
            self.extend(data)

    # Public Interface

    def set(self, key: str, value: str) -> None:
        # Store original case of key
        self._store[key.lower()] = [key, value]

    def get(self, key: str) -> str:
        return self._store[key.lower()][1]

    def getlist(self, key: str) -> List[str]:
        return self._store[key.lower()][1:]

    def remove(self, key: str):
        del self._store[key.lower()]

    def add(self, key: str, value: str):
        box = self._store.setdefault(key.lower(), [key])
        box.append(value)

    def extend(self, data: HttpHeaderStream) -> None:
        seq = (
            data.items()
            if isinstance(data, MutableMapping)
            else cast(Iterable[Tuple[str, str]], data)
        )
        for key, val in seq:
            self.add(key, val)

    def __contains__(self, key: str) -> bool:
        return key.lower() in self._store

    def count_keys(self) -> int:
        return len(self._store.keys())

    def count_items(self) -> int:
        return sum(1 for _ in self.items())

    def items(self):
        for items in self._store.values():
            original_key = items[0]
            for idx, item in enumerate(items):
                if idx > 0:
                    yield original_key, item

    def __repr__(self) -> str:
        return str(list(self.items()))



# pylint: disable=consider-using-f-string
from pprint import pprint  # pylint: disable=unused-import
from threading import Thread
import time
from urllib.parse import unquote, quote

from urllib3 import PoolManager
from urllib3.util.retry import Retry
from urllib3.response import HTTPResponse
import pytest

from test_server import (
    TestServer,
    WaitTimeoutError,
    TestServerError,
    Response,
    Request,
    RequestNotProcessed,
)
import test_server

from .util import fixture_global_server, fixture_server  # pylint: disable=unused-import

NETWORK_TIMEOUT = 1
SPECIFIC_TEST_PORT = 10100
pool = PoolManager()


def request(
    url, data=None, method=None, headers=None, fields=None, retries_redirect=10
) -> HTTPResponse:
    params = {
        "headers": headers,
        "timeout": NETWORK_TIMEOUT,
        "retries": Retry(
            total=None,
            connect=0,
            read=0,
            redirect=retries_redirect,
            other=0,
        ),
        "fields": fields,
    }
    if data:
        assert isinstance(data, bytes)
        params["body"] = data
    if not method:
        method = "POST" if (data or fields) else "GET"
    return pool.request(method, url, **params)


# WTF: urllib3 makes TWO requests :-/
# def test_non_ascii_header(server: TestServer) -> None:
#    server.add_response(Response(headers=[("z", server.get_url() + "фыва")]))
#    res = request(server.get_url(), retries_redirect=False)
#    print(res.headers)


def test_non_ascii_header(server: TestServer) -> None:
    server.add_response(
        Response(status=301, headers=[("Location", server.get_url(quote("фыва")))])
    )
    server.add_response(Response())
    request(server.get_url())
    assert quote("фыва") in server.get_request().path


def test_get(server: TestServer) -> None:
    valid_data = b"zorro"
    server.add_response(Response(data=valid_data))
    res = request(server.get_url())
    assert res.data == valid_data


def test_non_utf_request_data(server: TestServer) -> None:
    server.add_response(Response(data=b"abc"))
    res = request(url=server.get_url(), data="конь".encode("cp1251"))
    assert res.data == b"abc"
    assert server.get_request().data == "конь".encode("cp1251")


def test_request_client_ip(server: TestServer) -> None:
    server.add_response(Response())
    request(server.get_url())
    assert server.address == server.get_request().client_ip


def test_path(server: TestServer) -> None:
    server.add_response(Response())
    request(server.get_url("/foo?bar=1"))
    assert server.get_request().path == "/foo"
    assert server.get_request().args["bar"] == "1"


def test_post(server: TestServer) -> None:
    server.add_response(Response(data=b"abc"), method="post")
    res = request(server.get_url(), b"req-data")
    assert res.data == b"abc"
    assert server.get_request().data == b"req-data"


def test_response_once_specific_method(server: TestServer) -> None:
    server.add_response(Response(data=b"bar"), method="get")
    server.add_response(Response(data=b"foo"))
    assert request(server.get_url()).data == b"bar"


def test_request_headers(server: TestServer) -> None:
    server.add_response(Response())
    request(server.get_url(), headers={"Foo": "Bar"})
    assert server.get_request().headers.get("foo") == "Bar"


def test_response_once_reset_headers(server: TestServer) -> None:
    server.add_response(Response(headers=[("foo", "bar")]))
    server.reset()
    res = request(server.get_url())
    assert res.status == 555
    assert b"No response" in res.data


def test_method_sleep(server: TestServer) -> None:
    server.add_response(Response())
    delay = 0.3

    start = time.time()
    request(server.get_url())
    elapsed = time.time() - start
    assert elapsed <= delay

    server.add_response(Response(sleep=delay))
    start = time.time()
    request(server.get_url())
    elapsed = time.time() - start
    assert elapsed > delay


def test_request_done_after_start(server: TestServer) -> None:
    server = TestServer()
    try:
        server.start()
        assert not server.request_is_done()
    finally:
        server.stop()


def test_request_done(server: TestServer) -> None:
    assert not server.request_is_done()
    server.add_response(Response())
    request(server.get_url())
    assert server.request_is_done()


def test_wait_request(server: TestServer) -> None:
    server.add_response(Response(data=b"foo"))

    def worker():
        time.sleep(1)
        request(server.get_url("?method=test-wait-request"))

    th = Thread(target=worker)
    th.start()
    with pytest.raises(WaitTimeoutError):
        server.wait_request(0.5)
    server.wait_request(2)
    th.join()


def test_request_cookies(server: TestServer) -> None:
    server.add_response(Response())
    request(url=server.get_url(), headers={"Cookie": "foo=bar"})
    assert server.get_request().cookies["foo"].value == "bar"


def test_default_header_content_type(server: TestServer) -> None:
    server.add_response(Response())
    info = request(server.get_url())
    assert info.headers["content-type"] == "text/html; charset=utf-8"


def test_custom_header_content_type(server: TestServer) -> None:
    server.add_response(
        Response(headers=[("Content-Type", "text/html; charset=koi8-r")])
    )
    info = request(server.get_url())
    assert info.headers["content-type"] == "text/html; charset=koi8-r"


def test_default_header_server(server: TestServer) -> None:
    server.add_response(Response())
    info = request(server.get_url())
    assert info.headers["server"] == ("TestServer/%s" % test_server.__version__)


def test_custom_header_server(server: TestServer) -> None:
    server.add_response(Response(headers=[("Server", "Google")]))
    info = request(server.get_url())
    assert info.headers["server"] == "Google"


def test_options_method(server: TestServer) -> None:
    server.add_response(Response(data=b"abc"))
    res = request(url=server.get_url(), method="OPTIONS")
    assert server.get_request().method == "OPTIONS"
    assert res.data == b"abc"


def test_multiple_start_stop_cycles() -> None:
    for _ in range(30):
        server = TestServer()
        server.start()
        try:
            server.add_response(Response(data=b"zorro"), count=10)
            for _ in range(10):
                res = request(server.get_url())
                assert res.data == b"zorro"
        finally:
            server.stop()


def test_specific_port() -> None:
    server = TestServer(address="localhost", port=SPECIFIC_TEST_PORT)
    try:
        server.start()
        server.add_response(Response(data=b"abc"))
        data = request(server.get_url()).data
        assert data == b"abc"
    finally:
        server.stop()


def test_null_bytes(server: TestServer) -> None:
    server.add_response(
        Response(
            status=302,
            headers=[
                ("Location", server.get_url().rstrip("/") + "/\x00/"),
            ],
        )
    )
    server.add_response(Response(data=b"zzz"))
    res = request(server.get_url())
    assert res.data == b"zzz"
    assert unquote(server.get_request().path) == "/\x00/"


def test_callback(server: TestServer) -> None:
    def get_callback():
        return {
            "type": "response",
            "data": b"Hello",
            "headers": [
                ("method", "get"),
            ],
        }

    def post_callback():
        return {
            "type": "response",
            "status": 201,
            "data": b"hey",
            "headers": [
                ("method", "post"),
                ("set-cookie", "foo=bar"),
            ],
        }

    server.add_response(Response(callback=get_callback))
    server.add_response(Response(callback=post_callback), method="post")

    info = request(server.get_url())
    assert info.headers.get("method") == "get"
    assert info.data == b"Hello"

    info = request(server.get_url(), b"key=val")
    assert info.headers["method"] == "post"
    assert info.headers["set-cookie"] == "foo=bar"
    assert info.data == b"hey"
    assert info.status == 201


def test_response_data_invalid_type(server: TestServer) -> None:
    server.add_response(Response(data=1))  # type: ignore
    res = request(server.get_url())
    assert res.status == 555
    assert b"must be bytes" in res.data


def test_stop_not_started_server() -> None:
    server = TestServer()
    server.stop()


def test_start_request_stop_same_port() -> None:
    server = TestServer()
    for _ in range(10):
        try:
            server.start()
            server.add_response(Response())
            request(server.get_url())
        finally:
            server.stop()


def test_file_uploading(server: TestServer) -> None:
    server.add_response(Response())
    request(
        server.get_url(),
        fields={
            "image": ("emoji.png", b"zzz"),
        },
    )
    assert server.get_request().files["image"][0]["name"] == "image"


def test_callback_response_not_dict(server: TestServer) -> None:
    def callback():
        return ["foo", "bar"]

    server.add_response(Response(callback=callback))
    res = request(server.get_url())
    assert res.status == 555
    assert b"is not a dict" in res.data


def test_callback_response_invalid_type(server: TestServer) -> None:
    def callback():
        return {
            "foo": "bar",
        }

    server.add_response(Response(callback=callback))
    res = request(server.get_url())
    assert res.status == 555
    assert b"invalid type key" in res.data


def test_callback_response_invalid_key(server: TestServer) -> None:
    def callback():
        return {
            "type": "response",
            "foo": "bar",
        }

    server.add_response(Response(callback=callback))
    res = request(server.get_url())
    assert res.status == 555
    assert b"contains invalid key" in res.data


def test_callback_data_non_bytes(server: TestServer) -> None:
    def callback():
        return {
            "type": "response",
            "data": "bar",
        }

    server.add_response(Response(callback=callback))
    res = request(server.get_url())
    assert res.status == 555
    assert b"must be bytes" in res.data


def test_invalid_response_key() -> None:
    with pytest.raises(TypeError) as ex:
        # pylint: disable=unexpected-keyword-arg
        Response(foo="bar")  # type: ignore
    assert "unexpected keyword argument" in str(ex.value)


def test_get_request_no_request(server: TestServer) -> None:
    with pytest.raises(RequestNotProcessed):
        server.get_request()


def test_add_response_invalid_method(server: TestServer) -> None:
    with pytest.raises(TestServerError) as ex:
        server.add_response(Response(), method="foo")
    assert "Invalid method" in str(ex.value)


def test_add_response_count_minus_one(server: TestServer) -> None:
    server.add_response(Response(), count=-1)
    for _ in range(3):
        assert 200 == request(server.get_url()).status


def test_add_response_count_one_default(server: TestServer) -> None:
    server.add_response(Response())
    assert 200 == request(server.get_url()).status
    assert b"No response" in request(server.get_url()).data

    server.add_response(Response(), count=1)
    assert 200 == request(server.get_url()).status
    assert b"No response" in request(server.get_url()).data


def test_add_response_count_two(server: TestServer) -> None:
    server.add_response(Response(), count=2)
    assert 200 == request(server.get_url()).status
    assert 200 == request(server.get_url()).status
    assert b"No response" in request(server.get_url()).data


def test_raw_callback(server):
    def callback():
        return b"HTTP/1.0 200 OK\nFoo: Bar\nGaz: Baz\nContent-Length: 5\n\nhello"

    server.add_response(Response(raw_callback=callback))
    res = request(server.get_url())
    assert "foo" in res.headers
    assert b"hello" == res.data


def test_raw_callback_invalid_type(server):
    def callback():
        return "hey"

    server.add_response(Response(raw_callback=callback))
    res = request(server.get_url())
    assert b"must return bytes" in res.data


def test_request_property(server):
    server.add_response(Response())
    request(server.get_url())
    assert isinstance(server.request, Request)


def test_put_request(server):
    server.add_response(Response())
    request(server.get_url(), data=b"foo", method="put")
    assert server.request.method == "PUT"



from test_server.server import *  # noqa pylint: disable=wildcard-import
from test_server.structure import *  # noqa pylint: disable=wildcard-import
from test_server.error import *  # noqa pylint: disable=wildcard-import
from test_server.version import TEST_SERVER_VERSION

__version__ = TEST_SERVER_VERSION



from typing import Mapping, MutableMapping, Dict


def test_dict():
    box = {}
    assert isinstance(box, Mapping)
    assert isinstance(box, MutableMapping)
    assert isinstance(box, Dict)
    assert isinstance(box, dict)


def test_mapping_instance():
    class Box(MutableMapping):  # pragma: no cover
        def __delitem__(self, key):
            pass

        def __getitem__(self, key):
            pass

        def __setitem__(self, key, val):
            pass

        def __iter__(self):
            return iter([])

        def __len__(self):
            return 0

    box = Box()
    assert isinstance(box, Mapping)
    assert isinstance(box, MutableMapping)
    assert not isinstance(box, Dict)
    assert not isinstance(box, dict)



__all__ = [
    "TestServerError",
    "WaitTimeoutError",
    "InternalError",
    "RequestNotProcessed",
    "NoResponse",
]


class TestServerError(Exception):
    """Base class for all errrors which belogns to test_server package"""


class WaitTimeoutError(TestServerError):
    """Raised by wait_request method if it timed out waiting a request done"""


class InternalError(TestServerError):
    """Raised when exception happens during the processing request sent by client"""


class RequestNotProcessed(TestServerError):
    """Raised by get_request method when no request has been processed"""


class NoResponse(TestServerError):
    """Raised by get_response method when no response data is available
    to hande the request"""



# pylint: disable=consider-using-f-string
from pprint import pprint  # pylint: disable=unused-import
import time
from collections import defaultdict
from threading import Thread, Event
import cgi
from io import BytesIO
import logging
from typing import Optional, Callable, List, MutableMapping, cast

from socketserver import ThreadingMixIn, TCPServer
from http.server import BaseHTTPRequestHandler
from http.cookies import SimpleCookie
from urllib.parse import urljoin, parse_qsl

from .version import TEST_SERVER_VERSION
from .error import (
    TestServerError,
    WaitTimeoutError,
    InternalError,
    RequestNotProcessed,
    NoResponse,
)
from .structure import HttpHeaderStorage, HttpHeaderStream

__all__: list = ["TestServer", "WaitTimeoutError", "Response", "Request"]

INTERNAL_ERROR_RESPONSE_STATUS: int = 555


class HandlerResult:
    __slots__ = ["status", "headers", "data"]

    def __init__(
        self,
        status: Optional[int] = None,
        headers: Optional[HttpHeaderStorage] = None,
        data: Optional[bytes] = None,
    ) -> None:
        self.status = status if status is not None else 200
        self.headers = headers if headers else HttpHeaderStorage()
        self.data = data if data else b""


class Response(object):
    def __init__(
        self,
        callback: Optional[Callable] = None,
        raw_callback: Optional[Callable] = None,
        data: Optional[bytes] = None,
        headers: Optional[HttpHeaderStream] = None,
        sleep: Optional[float] = None,
        status: Optional[int] = None,
    ) -> None:
        self.callback = callback
        self.raw_callback = raw_callback
        self.data = b"" if data is None else data
        self.headers = HttpHeaderStorage(headers)
        self.sleep = sleep
        self.status = 200 if status is None else status


class Request(object):
    def __init__(
        self,
        args: Optional[dict] = None,
        client_ip: Optional[str] = None,
        cookies: Optional[SimpleCookie] = None,
        data: Optional[bytes] = None,
        files: Optional[dict] = None,
        headers: Optional[HttpHeaderStream] = None,
        method: Optional[str] = None,
        path: Optional[str] = None,
    ):
        self.args = {} if args is None else args
        self.client_ip = {} if client_ip is None else client_ip
        self.cookies: SimpleCookie = SimpleCookie() if cookies is None else cookies
        self.data = None if data is None else data
        self.files = {} if files is None else files
        self.headers = HttpHeaderStorage(headers)
        self.method = None if method is None else method
        self.path: str = "" if path is None else path


VALID_METHODS: List[str] = ["get", "post", "put", "delete", "options", "patch"]


class ThreadingTCPServer(ThreadingMixIn, TCPServer):
    allow_reuse_address: bool = True
    started: bool = False

    def __init__(
        self, server_address, RequestHandlerClass, test_server=None, **kwargs
    ) -> None:
        super().__init__(server_address, RequestHandlerClass, **kwargs)
        self.test_server = test_server
        self.test_server.server_started.set()


class TestServerHandler(BaseHTTPRequestHandler):
    server: ThreadingTCPServer

    def _collect_request_data(self, method: str) -> Request:
        data: MutableMapping = {
            "args": {},
            "headers": [],
            "files": defaultdict(list),
        }
        data["client_ip"] = self.client_address[0]
        try:
            qs = self.path.split("?")[1]
        except IndexError:
            qs = ""
        params = dict(parse_qsl(qs))
        for key, val in params.items():
            data["args"][key] = val
        for key, val in self.headers.items():
            data["headers"].append((key, val))

        path = self.path
        data["path"] = path.split("?")[0]
        data["method"] = method.upper()

        data["cookies"] = SimpleCookie(self.headers["Cookie"])

        clen = int(self.headers["Content-Length"] or "0")
        request_data = self.rfile.read(clen)
        data["data"] = request_data

        ctype = self.headers["Content-Type"]
        if ctype and ctype.split(";")[0] == "multipart/form-data":
            form = cgi.FieldStorage(
                fp=BytesIO(request_data),
                headers=cast(MutableMapping, self.headers),
                environ={
                    "REQUEST_METHOD": "POST",
                    "CONTENT_TYPE": self.headers["Content-Type"],
                },
            )
            for field_key in form.keys():  # pylint: disable=consider-using-dict-items
                box = form[field_key]
                for field in box if isinstance(box, list) else [box]:
                    data["files"].setdefault(field_key, []).append(
                        {
                            "name": field_key,
                            # "raw_filename": None,
                            "content_type": field.type,
                            "filename": field.filename,
                            "content": field.file.read(),
                        }
                    )

        return Request(**data)

    def _request_handler(self) -> None:
        try:
            test_srv = self.server.test_server  # pytype: disable=attribute-error
            method = self.command.lower()
            resp = test_srv.get_response(method)
            if resp.sleep:
                time.sleep(resp.sleep)
            test_srv.add_request(self._collect_request_data(method))

            result = HandlerResult()

            if resp.raw_callback:
                data = resp.raw_callback()
                if isinstance(data, bytes):
                    self.write_raw_response_data(data)
                    return
                else:
                    raise InternalError("Raw callback must return bytes data")

            if resp.callback:
                cb_res = resp.callback()
                if not isinstance(cb_res, dict):
                    raise InternalError("Callback response is not a dict")
                elif cb_res.get("type") == "response":
                    for key in cb_res:
                        if key not in (
                            "type",
                            "status",
                            "headers",
                            "data",
                        ):
                            raise InternalError(
                                "Callback response contains invalid key: %s" % key
                            )
                    if "status" in cb_res:
                        result.status = cb_res["status"]
                    if "headers" in cb_res:
                        result.headers.extend(cb_res["headers"])
                    if "data" in cb_res:
                        if isinstance(cb_res["data"], bytes):
                            result.data = cb_res["data"]
                        else:
                            raise InternalError(
                                'Callback repsponse field "data" must be bytes'
                            )
                else:
                    raise InternalError(
                        "Callback response has invalid type key: %s"
                        % cb_res.get("type", "NA")
                    )
            else:
                result.status = resp.status
                result.headers.extend(resp.headers.items())
                data = resp.data
                if isinstance(data, bytes):
                    result.data = data
                else:
                    raise InternalError('Response parameter "data" must be bytes')

            port = self.server.test_server.port  # pytype: disable=attribute-error
            result.headers.set("Listen-Port", str(port))
            if "content-type" not in result.headers:
                result.headers.set("Content-Type", "text/html; charset=utf-8")
            if "server" not in result.headers:
                result.headers.set("Server", "TestServer/%s" % TEST_SERVER_VERSION)

            self.write_response_data(result.status, result.headers, result.data)
        except Exception as ex:  # pylint: disable=broad-except
            logging.exception("Unexpected error happend in test server request handler")
            self.write_response_data(
                INTERNAL_ERROR_RESPONSE_STATUS,
                HttpHeaderStorage(),
                str(ex).encode("utf-8"),
            )
        finally:
            test_srv.num_req_processed += 1

    def write_response_data(
        self, status: int, headers: HttpHeaderStorage, data: bytes
    ) -> None:
        self.send_response(status)
        for key, val in headers.items():
            self.send_header(key, val)
        self.end_headers()
        self.wfile.write(data)

    def write_raw_response_data(self, data: bytes) -> None:
        self.wfile.write(data)
        # pylint: disable=attribute-defined-outside-init
        self._headers_buffer: List[str] = []

    # https://github.com/python/cpython/blob/main/Lib/http/server.py
    def send_response(self, code: int, message: Optional[str] = None) -> None:
        """
        Custom method which does not send Server and Date headers

        This method overrides standard method from super class.
        """
        self.log_request(code)
        self.send_response_only(code, message)

    do_GET = _request_handler
    do_POST = _request_handler
    do_PUT = _request_handler
    do_DELETE = _request_handler
    do_OPTIONS = _request_handler
    do_PATCH = _request_handler


class TestServer(object):
    def __init__(self, address="127.0.0.1", port=0) -> None:
        self.server_started: Event = Event()
        self._requests: List = []
        self._responses: MutableMapping = defaultdict(list)
        self.port: Optional[int] = None
        self._config_port: int = port
        self.address: str = address
        self._thread: Optional[Thread] = None
        self._server: Optional[ThreadingTCPServer] = None
        self._started: Event = Event()
        self.num_req_processed: int = 0
        self.reset()

    def _thread_server(self) -> None:
        """Ask HTTP server start processing requests

        This function is supposed to be run in separate thread.
        """

        self._server = ThreadingTCPServer(
            (self.address, self._config_port), TestServerHandler, test_server=self
        )
        self._server.serve_forever(poll_interval=0.1)

    # ****************
    # Public Interface
    # ****************

    def add_request(self, req: Request) -> None:
        self._requests.append(req)

    def reset(self) -> None:
        self.num_req_processed = 0
        self._requests.clear()
        self._responses.clear()

    def start(self, daemon: bool = True) -> None:
        """Start the HTTP server."""
        self._thread = Thread(
            target=self._thread_server,
        )
        self._thread.daemon = daemon
        self._thread.start()
        self.wait_server_started()
        self.port = cast(ThreadingTCPServer, self._server).socket.getsockname()[1]

    def wait_server_started(self) -> None:
        # I could not foind another way
        # to handle multiple socket issues
        # other than taking some sleep
        time.sleep(0.01)
        self.server_started.wait()

    def stop(self) -> None:
        if self._server:
            self._server.shutdown()
            self._server.server_close()

    def get_url(self, path: str = "", port: Optional[int] = None) -> str:
        """Build URL that is served by HTTP server."""
        if port is None:
            port = cast(int, self.port)
        return urljoin("http://%s:%d" % (self.address, port), path)

    def wait_request(self, timeout: float) -> None:
        """Stupid implementation that eats CPU."""
        start: float = time.time()
        while True:
            if self.num_req_processed:
                break
            time.sleep(0.01)
            if time.time() - start > timeout:
                raise WaitTimeoutError("No request processed in %d seconds" % timeout)

    def request_is_done(self) -> bool:
        return self.num_req_processed > 0

    def get_request(self) -> Request:
        try:
            return self._requests[-1]
        except IndexError as ex:
            raise RequestNotProcessed("Request has not been processed") from ex

    @property
    def request(self) -> Request:
        return self.get_request()

    def add_response(
        self, resp: Response, count: int = 1, method: Optional[str] = None
    ) -> None:
        assert method is None or isinstance(method, str)
        assert count < 0 or count > 0
        if method and method not in VALID_METHODS:
            raise TestServerError("Invalid method: %s" % method)
        self._responses[method].append(
            {
                "count": count,
                "response": resp,
            },
        )

    def get_response(self, method: str) -> Response:
        while True:
            item = None
            scope = None
            try:
                scope = self._responses[method]
                item = scope[0]
            except IndexError:
                try:
                    scope = self._responses[None]
                    item = scope[0]
                except IndexError as ex:
                    raise NoResponse("No response available") from ex
            if item["count"] == -1:
                return item["response"]
            else:
                item["count"] -= 1
                if item["count"] < 1:
                    scope.pop(0)
                return item["response"]



import pytest

from test_server import TestServer

STATE = {"server": None}


@pytest.fixture(scope="session", name="global_server")
def fixture_global_server():
    if not STATE["server"]:
        srv = TestServer()
        srv.start()
        STATE["server"] = srv
    yield STATE["server"]
    if STATE["server"]:
        STATE["server"].stop()
        STATE["server"] = None


@pytest.fixture(scope="function", name="server")
def fixture_server(global_server):
    global_server.reset()
    return global_server



import os

from setuptools import setup

ROOT = os.path.dirname(os.path.realpath(__file__))

setup(
    # Meta data
    name="test_server",
    version="0.0.40",
    author="Gregory Petukhov",
    author_email="lorien@lorien.name",
    maintainer="Gregory Petukhov",
    maintainer_email="lorien@lorien.name",
    url="https://github.com/lorien/test_server",
    description="Server for testing HTTP clients",
    long_description=open(os.path.join(ROOT, "README.rst"), encoding="utf-8").read(),
    download_url="https://pypi.python.org/pypi/test_server",
    keywords="test testing server http-server",
    license="MIT License",
    # Package files
    packages=["test_server"],
    package_data={"test_server": ["py.typed"]},
    zip_safe=False,
    install_requires=[],
    # Topics
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import pytest

from test_server.structure import HttpHeaderStorage


def test_constructor_no_data():
    HttpHeaderStorage()


def test_constructor_dict():
    HttpHeaderStorage({"foo": "bar"})


def test_constructor_list():
    HttpHeaderStorage([("foo", "bar")])


def test_set_get_simple_value():
    obj = HttpHeaderStorage()
    obj.set("foo", "bar")
    assert obj.get("foo") == "bar"


def test_set_get_multi_value():
    obj = HttpHeaderStorage()
    obj.add("foo", "bar")
    obj.add("foo", "baz")
    assert obj.getlist("foo") == ["bar", "baz"]


def test_delitem():
    obj = HttpHeaderStorage()
    with pytest.raises(KeyError):
        obj.remove("foo")
    obj.set("foo", "bar")
    obj.remove("foo")
    with pytest.raises(KeyError):
        obj.remove("foo")


def test_repr():
    obj = HttpHeaderStorage()
    obj.add("foo", "bar")
    obj.add("foo", "baz")
    assert repr(obj) == "[('foo', 'bar'), ('foo', 'baz')]"


def test_constructor_key_multivalue():
    obj = HttpHeaderStorage([("set-cookie", "foo=bar"), ("set-cookie", "baz=gaz")])
    assert obj.getlist("set-cookie") == ["foo=bar", "baz=gaz"]


def test_count_keys():
    obj = HttpHeaderStorage()
    obj.add("foo", "bar")
    obj.add("foo", "baz")
    assert 1 == obj.count_keys()


def test_count_items():
    obj = HttpHeaderStorage()
    obj.add("foo", "bar")
    obj.add("foo", "baz")
    assert 2 == obj.count_items()
